/*
 * Grundkurs Informatik - Beispielprogramm zur binären Suche
 * H. Ernst, J. Schmidt 2016
 *
 * Vergleich der binären Suche und der Interpolationssuche
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define MAX 30000

void sort_insb(long int a[], long int n);

long int srch_int(long int x, long int a[], long int n, int *count);

long int srch_bin(long int x, long int a[], long int n, int *count);


//------------------------------------------------------------------------
// Vergleich der binären Suche und der Interpolationssuche durch Suche
// in einem mit Zufallszahlen vorbesetzten und danach geordneten Feld.
//------------------------------------------------------------------------
int main() {
    long int i, k, a[MAX], n = MAX, m = 100;
    int count = 0;  // Schrittzähler
    float sum;
    time_t t; // Zeit

    printf("\nVergleich von Suchalgorithmen");
    printf("\n=============================");
    srand((unsigned) time(NULL)); // init. Zufallszahlengenerator

    for (i = 0; i < MAX; i++) a[i] = rand(); // mit Zufallszahlen vorbesetzen

    printf("\n\nSortieren ... ");
    t = clock();
    sort_insb(a, n);
    printf("beendet nach %5.3f sec", // benötigte Zeit
           (float) (difftime(clock(), t) / CLOCKS_PER_SEC));

    printf("\n\nBinaere Suche:"); // Binäre Suche
    t = clock();
    sum = 0.0;
    for (k = 0; k < m * n; k++) { // mittlere Anzahl der Schritte
        srch_bin(rand(), a, n, &count);
        sum += count;
    }
    printf(" %6.3f sec", (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printf("\nMittlere Anzahl der Schritte: %5.2f", sum / m / n);

    printf("\n\nInterpolations-Suche:"); // Interpolationssuche
    t = clock();
    sum = 0.0;
    for (k = 0; k < m * n; k++) { // mittlere Anzahl der Schritte
        srch_int(rand(), a, n, &count);
        sum += count;
    }
    printf(" %6.3f msec", (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printf("\nMittlere Anzahl der Schritte: %5.2f", sum / m / n);

    return EXIT_SUCCESS;
}

//------------------------------------------------------------------------
// Binäre Suche nach einem Element x in einem aufsteigend
// geordneten Integer-Feld a der Dimension n.
// Rückgabewert: Index des gefundenen Elements
// oder -1, wenn nicht gefunden.
//------------------------------------------------------------------------
long int srch_bin(long int x, long int a[], long int n, int *count) {
    long int anf = 0, end = n - 1, m = 0; // Intervall
    *count = 0; // Schrittzähler

    while (anf <= end && x != a[m]) { // Solange Element x nicht gefunden
        m = (anf + end) / 2; // halbiere Intervall
        if (x < a[m]) end = m - 1; // suche im linken Intervall
        else anf = m + 1; // suche im rechten Intervall
        (*count)++;
    }

    if (a[m] == x) return m; // Element gefunden

    return -1; // Element nicht gefunden
}

//------------------------------------------------------------------------
// Interpolationssuche nach einem Element x in einem aufsteigend
// geordneten Integer-Feld a der Dimension n.
// Rückgabewert: Index des gefundenen Elements
// oder -1, wenn nicht gefunden.
//------------------------------------------------------------------------
long int srch_int(long int x, long int a[], long int n, int *count) {
    long int anf = 0, end = n - 1, m = 0; // Intervall
    *count = 0; // Schrittzähler

    if (x < a[anf] || x > a[end])
        return -1; // Element zu klein oder zu groß

    while (anf < end) { // solange Element x nicht gefunden
        m = anf + (end - anf) * (x - a[anf] + 1) / (a[end] - a[anf] + 1);
        if (m < anf) m = anf; // Bereichsgrenzen prüfen
        else if (m > end) m = end;
        if (x < a[m]) end = m - 1; // suche im linken Intervall
        else anf = m + 1; // suche im rechten Intervall
        (*count)++;
    }

    if (a[m] == x) return m; // Element gefunden
    if (a[anf] == x) return anf; // Element gefunden (Sonderfall 1)
    if (a[end] == x) return end; // Element gefunden (Sonderfall 2)

    return -1;
}

//------------------------------------------------------------------------
// Sortieren eines Int-Arrays a mit Dimension n in aufsteigende Ordnung
// durch direktes Einfügen mit binärer Suche der Einfügestelle.
//------------------------------------------------------------------------
void sort_insb(long int a[], long int n) {
    long int i, j, ug, og, x;

    for (i = 1; i < n; i++) { // Schleife durch das Array
        x = a[i];
        ug = 0;
        og = i - 1; // Grenzen des sortierten Bereichs
        while (ug <= og) { // Binäre Suche der Einfügestelle
            j = (ug + og) / 2;
            if (x < a[j]) og = j - 1;
            else ug = j + 1;
        }

        for (j = i; j > ug; j--)
            a[j] = a[j - 1]; // Elemente verschieben
        a[ug] = x; // x einfügen
    }

    return;
}